import math
from typing import Optional

import numpy as np


# modified from https://stackoverflow.com/questions/34372480/
def rotate_points(px, py, origin, angle):
    """
    Rotate points counterclockwise by a given angle around a given origin.

    The angle should be given in degrees.
    """
    angle = math.radians(angle)
    ox, oy = origin

    qx = ox + np.cos(angle) * (px - ox) - np.sin(angle) * (py - oy)
    qy = oy + np.sin(angle) * (px - ox) + np.cos(angle) * (py - oy)
    return qx, qy


def normalize(arr: np.ndarray,
              vmin: Optional[float] = None,
              vmax: Optional[float] = None,
              ):
    if (vmin is None) & (vmax is None):
        return arr

    amin = np.amin(arr)
    amax = np.amax(arr)
    narr = (arr - amin) / (amax - amin)
    vmin = amin if vmin is None else vmin
    vmax = amax if vmax is None else vmax
    return narr * (vmax - vmin) + vmin



from __future__ import annotations

from typing import List, Tuple, Any, Dict

import matplotlib as mpl
import matplotlib.axes
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import PatchCollection, LineCollection
from mpl_toolkits.mplot3d import Axes3D

from legendkit import CatLegend, Colorbar
from milkviz.utils import set_ticks, set_spines, \
    color_mapper_cat, rotate_points, color_mapper_val, doc, create_cmap, normalize, \
    set_default


def _init_canvas(ax):
    """For map function only, clear all axis"""
    set_ticks(ax)
    ax.set_aspect("equal")
    ax.set_xticklabels([])
    ax.set_yticklabels([])


def _color_array(c_array, cmap, cmapper, types):
    color_array = None if cmapper is None else [cmapper[t] for t in types]
    if cmapper is None:
        if c_array is not None:
            cmapper = color_mapper_cat(types, c_array=c_array)
        else:
            cmapper = color_mapper_cat(types, cmap=cmap)
        color_array = [cmapper[t] for t in types]

    return cmapper, color_array


def _set_cat_legend(cmapper, ax, legend_kw):
    labels, colors = zip(*cmapper.items())
    legend_options = dict(
        handle="circle",
        bbox_to_anchor=(1.05, 0.5),
        bbox_transform=ax.transAxes,
        loc="center left",
        title_align="left",
    )
    legend_options = {**legend_options, **legend_kw}
    CatLegend(colors, labels, ax=ax, **legend_options)


def _set_cbar(values, cmap, ax, cbar_kw):
    cmin = np.nanmin(values)
    cmax = np.nanmax(values)
    cbar_options = dict(
        bbox_to_anchor=(1.05, 0.5),
        bbox_transform=ax.transAxes,
        loc="center left",
        title_align="left",
    )
    cbar_options = {**cbar_options, **cbar_kw}
    Colorbar(vmin=cmin, vmax=cmax, cmap=cmap, ax=ax, **cbar_options)


@doc
def point_map(
        x,
        y,
        types: List | np.ndarray = None,
        types_colors: Dict = None,
        values: List | np.ndarray = None,
        links: List | np.ndarray = None,
        vmin: float = None,
        vmax: float = None,
        colors: List = None,
        cmap: Any = None,
        rotate: int = None,
        markersize: int = 5,
        linecolor: Any = "#cccccc",
        linewidth: int = 1,
        no_spines: bool = True,
        legend: bool = True,
        legend_kw: Dict = None,
        cbar_kw: Dict = None,
        ax: mpl.axes.Axes = None,
):
    """Point map

    Args:
        x: The x-coord array
        y: The y-coord array
        types: [types] of points
        types_colors: A dictionary that tells color for every type, key is the type and value is the color
        values: [values] of points
        vmin, vmax: [vminmax]
        links: The links between points, should be a list of (point_index_1, point_index_2)
        colors: [hue]
        cmap: [cmap]
        rotate: The degree to rotate the whole plot according to origin
        markersize: The size of marker
        linecolor: The color of lines
        linewidth: The width of lines
        no_spines: [no_spines]
        legend: [legend]
        legend_kw: [legend_kw]
        cbar_kw: [cbar_kw]
        ax: [ax]

    Returns:
        [return_obj]

    """
    if ax is None:
        ax = plt.gca()
    if no_spines:
        set_spines(ax)
    legend_kw = set_default(legend_kw, {})
    cbar_kw = set_default(cbar_kw, {})

    _init_canvas(ax)

    if rotate is not None:
        x, y = rotate_points(x, y, (0, 0), rotate)

    if links is not None:
        if not isinstance(linecolor, str):
            if len(linecolor) != len(links):
                raise ValueError("Length of linecolor must match to links")
        lines = np.array([[[x[i1], y[i1]], [x[i2], y[i2]]] for i1, i2 in links])
        line_collections = LineCollection(lines, linewidths=linewidth, edgecolors=linecolor, zorder=-100)
        ax.add_collection(line_collections)

    if types is not None:
        cmap = set_default(cmap, "echarts")
        cmapper, color_array = _color_array(colors, cmap, types_colors, types)

        ax.scatter(x=x, y=y, c=color_array, s=markersize)
        if legend:
            _set_cat_legend(cmapper, ax, legend_kw)
    else:
        if values is not None:
            cmap = set_default(cmap, "OrRd")
            values = normalize(values, vmin=vmin, vmax=vmax)
            if colors is not None:
                vc_mapper = dict(zip(values, colors))
                cmap = create_cmap([vc_mapper[v] for v in sorted(values)])
            ax.scatter(x=x, y=y, c=values, s=markersize, cmap=cmap)
            if legend:
                _set_cbar(values, cmap, ax, cbar_kw)
        else:
            ax.scatter(x=x, y=y, s=markersize)

    return ax


@doc
def point_map3d(
        x,
        y,
        z,
        types: List | np.ndarray = None,
        types_colors: Dict = None,
        values: List | np.ndarray = None,
        vmin: float = None,
        vmax: float = None,
        colors: List | np.ndarray = None,
        cmap: Any = None,
        legend: bool = True,
        legend_kw: Dict = None,
        cbar_kw: Dict = None,
        markersize: int = 5,
        ax: mpl.axes.Axes = None,
):
    """Point map in 3D

        Args:
            x: The x-coord array
            y: The y-coord array
            z: The z-coord array
            types: [types] of points
            types_colors: A dictionary that tells color for every type, key is the type and value is the color
            values: [values] of points
            vmin, vmax: [vminmax]
            colors: [hue]
            cmap: [cmap]
            legend: [legend]
            legend_kw: [legend_kw]
            cbar_kw: [cbar_kw]
            markersize: The size of marker
            ax: [ax]

        Returns:
            [return_obj]

        """
    legend_kw = set_default(legend_kw, {})
    cbar_kw = set_default(cbar_kw, {})

    if ax is None:
        fig = plt.gcf()
        ax = fig.add_subplot(projection='3d')
    else:
        if not isinstance(ax, Axes3D):
            raise TypeError(f"If you want to use your own axes, initialize it as 3D")
    ax.set_xticklabels([])
    ax.set_yticklabels([])
    ax.set_zticklabels([])
    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")

    if types is not None:
        cmap = set_default(cmap, "tab20")
        cmapper, color_array = _color_array(colors, cmap, types_colors, types)

        ax.scatter(x, y, z, c=color_array, s=markersize)
        if legend:
            # bbox: 1.2, -0.1
            _set_cat_legend(cmapper, ax, legend_kw)
    else:
        if values is not None:
            cmap = set_default(cmap, "OrRd")
            values = normalize(values, vmin, vmax)
            if colors is not None:
                vc_mapper = dict(zip(values, colors))
                cmap = create_cmap([vc_mapper[v] for v in sorted(values)])
            ax.scatter(x, y, z, c=values, s=markersize, cmap=cmap)
            if legend:
                # pos: 1.2, 0.05
                _set_cbar(values, cmap, ax, cbar_kw)
        else:
            ax.scatter(x, y, z, s=markersize)

    return ax


@doc
def polygon_map(
        polygons: List[List[Tuple[float, float]]],
        types: List | np.ndarray = None,
        types_colors: Dict = None,
        values: List | np.ndarray = None,
        vmin: float = None,
        vmax: float = None,
        colors: List = None,
        cmap: Any = None,
        legend: bool = True,
        legend_kw: Dict = None,
        cbar_kw: Dict = None,
        rotate: int = None,
        no_spines: bool = True,
        ax: mpl.axes.Axes = None,
):
    """Polygon map

    Args:
        polygons: A list of polygons, a polygon is represented by a list of points
        types: [types] of polygons
        types_colors: A dictionary that tells color for every type, key is the type and value is the color
        values: [values] of polygons
        vmin, vmax: [vminmax]
        colors: [hue]
        cmap: [cmap]
        legend: [legend]
        legend_kw: [legend_kw]
        cbar_kw: [cbar_kw]
        rotate: The degree to rotate the whole plot according to origin
        no_spines: [no_spines]
        ax: [ax]

    Returns:
        [return_obj]

    """
    polygons = [np.array(polygon) for polygon in polygons]
    vstack_poly = np.vstack(polygons)
    xmin, ymin = np.min(vstack_poly, axis=0)
    xmax, ymax = np.max(vstack_poly, axis=0)
    legend_kw = set_default(legend_kw, {})
    cbar_kw = set_default(cbar_kw, {})

    if rotate is not None:
        rotated_polygons = []
        for polygon in polygons:
            x, y = rotate_points(polygon[:, 0], polygon[:, 1], (0, 0), rotate)
            rotated_polygons.append([(i, j) for i, j in zip(x, y)])
        polygons = rotated_polygons

    if ax is None:
        ax = plt.gca()
    if no_spines:
        set_spines(ax)

    _init_canvas(ax)

    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)

    if types is not None:
        cmap = set_default(cmap, "echarts")
        cmapper, color_array = _color_array(colors, cmap, types_colors, types)
        patches = [mpatches.Polygon(polygon) for polygon in polygons]
        patches_collections = PatchCollection(patches, facecolors=[cmapper[t] for t in types])
        ax.add_collection(patches_collections)
        if legend:
            _set_cat_legend(cmapper, ax, legend_kw)
    else:
        if values is not None:
            cmap = set_default(cmap, "OrRd")
            values = normalize(values, vmin, vmax)
            if colors is not None:
                vc_mapper = dict(zip(values, colors))
                cmap = create_cmap([vc_mapper[v] for v in sorted(values)])
            colors = color_mapper_val(values, cmap=cmap)
            patches = [mpatches.Polygon(polygon, color=c) for polygon, c in zip(polygons, colors)]
            patches_collections = PatchCollection(patches, facecolors=colors, cmap=cmap)
            ax.add_collection(patches_collections)
            if legend:
                _set_cbar(values, cmap, ax, cbar_kw)
        else:
            patches = [mpatches.Polygon(polygon) for polygon in polygons]
            patches_collections = PatchCollection(patches)
            ax.add_collection(patches_collections)

    return ax



from __future__ import annotations

from typing import Literal, Optional, List, Union, Tuple, Any, Dict

import matplotlib as mpl
import matplotlib.axes
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.cm import ScalarMappable
from matplotlib.collections import PatchCollection
from matplotlib.colors import Colormap, Normalize, is_color_like

from legendkit import SizeLegend, Colorbar, EllipseColorbar
from milkviz.utils import doc, norm_arr, set_default, set_spines, set_ticks

DotPatch = Literal["circle", "pie"]


class DotHeatmap:

    def __init__(
            self,
            dot_size: np.ndarray,
            dot_hue: np.ndarray = None,
            matrix_hue: np.ndarray = None,
            xticklabels: List[str] = None,
            yticklabels: List[str] = None,
            xlabel: str = None,
            ylabel: str = None,
            sizes: Tuple[int, int] = (1, 200),
            dtype: Any = None,
            dot_patch: DotPatch = "circle",
            dot_outline: bool = True,
            outline_color: Any = "#999999",
            alpha: float = 1.0,
            dot_cmap: Union[str, Colormap, None] = None,
            matrix_cmap: Union[str, Colormap, None] = None,
            dot_size_legend_kw: Dict = None,
            dot_hue_cbar_kw: Dict = None,
            matrix_cbar_kw: Dict = None,
            no_spines: bool = True,
            no_ticks: bool = True,
            ax: Optional[mpl.axes.Axes] = None,
    ) -> None:
        self.ax = ax
        self.dot_size = dot_size
        self.dot_hue = dot_hue
        self.matrix_hue = matrix_hue
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.xticklabels = xticklabels if xticklabels is not None else []
        self.yticklabels = yticklabels if yticklabels is not None else []
        self.sizes = sizes
        self.dtype = dtype
        self.no_spines = no_spines
        self.no_ticks = no_ticks

        self.dot_cmap = set_default(dot_cmap, "RdBu")
        self.matrix_cmap = set_default(matrix_cmap, "YlGn")

        self.dot_size_legend = None
        self.dot_hue_cbar = None
        self.matrix_cbar = None

        self.dot_size_legend_kw = set_default(dot_size_legend_kw, {})
        self.dot_hue_cbar_kw = set_default(dot_hue_cbar_kw, {})
        self.matrix_cbar_kw = set_default(matrix_cbar_kw, {})

        # init the grid
        self._set_grid()

        if dot_hue is not None:
            if isinstance(dot_hue, str):
                circ_colors = np.repeat(dot_hue, dot_size.size)
            else:
                if np.array((dot_hue.shape == dot_size.shape)).all():
                    c_array = dot_hue.flatten()
                    if is_color_like(c_array[0]):
                        circ_colors = c_array
                    else:
                        cmin = np.nanmin(c_array)
                        cmax = np.nanmax(c_array)
                        mapper = ScalarMappable(Normalize(vmin=cmin, vmax=cmax), cmap=self.dot_cmap)
                        circ_colors = mapper.to_rgba(c_array)
                        self.add_dot_hue_cbar(cmin, cmax)
                else:
                    raise ValueError("dot_hue does not match the shape of dot_size")
        else:
            circ_colors = np.repeat("#D75455", dot_size.size)

        if dot_patch == "circle":
            circ_size = norm_arr(dot_size, sizes)
            self.ax.scatter(self.xcoord, self.ycoord, s=circ_size, c=circ_colors)
            self.add_dot_size_legend()
        else:
            norm_sizes = norm_arr(dot_size, size=(0, 1))
            for rx, ry, ratio, c in zip(self.xcoord, self.ycoord, norm_sizes, circ_colors):
                if not np.isnan(ratio):
                    pie = mpatches.Wedge((rx, ry), 0.4, 90, 360 * ratio + 90, edgecolor=outline_color, facecolor=c, lw=0.5,
                                         alpha=alpha)
                    if dot_outline:
                        pie_bg = mpatches.Circle((rx, ry), 0.4, edgecolor=outline_color, facecolor="none", lw=0.5,
                                                 alpha=alpha)
                        self.ax.add_artist(pie_bg)
                    self.ax.add_artist(pie)

        if matrix_hue is not None:
            rects = [mpatches.Rectangle((rx - 0.5, ry - 0.5), 1, 1) \
                     for rx, ry in zip(self.xcoord, self.ycoord)]
            rects_colors = matrix_hue.flatten()
            rects_cmin = np.nanmin(rects_colors)
            rects_cmax = np.nanmax(rects_colors)
            rects = PatchCollection(rects, array=rects_colors, cmap=self.matrix_cmap, alpha=alpha)
            rects.set_clim(rects_cmin, rects_cmax)
            rects.set_zorder(-1)
            self.ax.add_collection(rects)

            self.add_matrix_cbar(rects)

    def _set_grid(self):
        Y, X = self.dot_size.shape
        x, y = np.meshgrid(np.arange(X), np.arange(Y))  # Get the coordinates
        self.xcoord = x.flatten()
        self.ycoord = y.flatten()

        if self.ax is None:
            figsize = (0.5 * X, 0.5 * Y)
            _, self.ax = plt.subplots(figsize=figsize)
        if self.no_spines:
            set_spines(self.ax)
        if self.no_ticks:
            set_ticks(self.ax)
        self.ax.set_aspect("equal")
        self.ax.set_xlim(-0.5, X - 0.5)
        self.ax.set_ylim(-0.5, Y - 0.5)
        self.ax.set(xlabel=self.xlabel, ylabel=self.ylabel)
        self.ax.set_xticks(np.arange(X), labels=self.xticklabels, rotation=90, ha="center")
        self.ax.set_yticks(np.arange(Y), labels=self.yticklabels)

    def add_dot_size_legend(self):
        options = dict(
            bbox_to_anchor=(1.05, 1),
            loc="upper left",
            bbox_transform=self.ax.transAxes,
            title_align="left"
        )
        options = {**options, **self.dot_size_legend_kw}
        self.dot_size_legend = SizeLegend(self.sizes,
                                          ax=self.ax,
                                          array=self.dot_size,
                                          dtype=self.dtype,
                                          **options
                                          )

    def add_dot_hue_cbar(self, vmin, vmax):
        options = dict(
            orientation="vertical",
            bbox_to_anchor=(1.05, 0),
            loc="lower left",
            bbox_transform=self.ax.transAxes,
            title_align="left"
        )
        options = {**options, **self.dot_hue_cbar_kw}
        self.dot_hue_cbar = EllipseColorbar(vmin=vmin,
                                            vmax=vmax,
                                            cmap=self.dot_cmap,
                                            ax=self.ax,
                                            **options
                                            )

    def add_matrix_cbar(self, cmapping):
        options = dict(
            orientation="horizontal",
            ticklocation="top",
            loc="lower left",
            bbox_to_anchor=(0, 1.05),
            bbox_transform=self.ax.transAxes,
            title_align="left"
        )
        options = {**options, **self.matrix_cbar_kw}
        self.matrix_cbar = Colorbar(cmapping,
                                    ax=self.ax,
                                    **options
                                    )


@doc
def dot_heatmap(
        dot_size: np.ndarray,
        dot_hue: np.ndarray = None,
        matrix_hue: np.ndarray = None,
        xticklabels: List[str] = None,
        yticklabels: List[str] = None,
        xlabel: str = None,
        ylabel: str = None,
        sizes: Tuple[int, int] = (1, 200),
        dtype: Any = None,
        dot_patch: DotPatch = "circle",
        dot_outline: bool = True,
        outline_color: Any = "#999999",
        alpha: float = 1.0,
        dot_cmap: Union[str, Colormap, None] = None,
        matrix_cmap: Union[str, Colormap, None] = None,
        dot_size_legend_kw: Dict = None,
        dot_hue_cbar_kw: Dict = None,
        matrix_cbar_kw: Dict = None,
        no_spines: bool = True,
        no_ticks: bool = True,
        ax: mpl.axes.Axes = None,
) -> DotHeatmap:
    """Dot heatmap + Matrix heatmap

    Args:
        dot_size: [size]
        dot_hue: [hue]
        matrix_hue: The color array that map to matrix colors
        xticklabels: [xticklabels]
        yticklabels: [yticklabels]
        xlabel: [xlabel]
        ylabel: [ylabel]
        sizes: [sizes]
        dtype: [dtype]
        dot_patch: The dot will be drawn as "circle" or "pie"
        dot_outline: To draw the dot outline, Default: True
        outline_color: The color of the dot outline, Default: "#999999"
        alpha: Control the opacity
        dot_cmap: The colormap for dot, default: "RdBu"
        matrix_cmap: The colormap for matrix, default: "YlGn"
        dot_size_legend_kw:
        dot_hue_cbar_kw:
        matrix_cbar_kw:
        no_spines: [no_spines]
        no_ticks: [no_ticks]
        ax: [ax]

    Returns:
        Return a `DotHeatmap` object

    """

    return DotHeatmap(
        dot_size,
        dot_hue,
        matrix_hue=matrix_hue,
        xticklabels=xticklabels,
        yticklabels=yticklabels,
        xlabel=xlabel,
        ylabel=ylabel,
        sizes=sizes,
        dtype=dtype,
        dot_patch=dot_patch,
        dot_outline=dot_outline,
        outline_color=outline_color,
        alpha=alpha,
        dot_cmap=dot_cmap,
        matrix_cmap=matrix_cmap,
        dot_size_legend_kw=dot_size_legend_kw,
        dot_hue_cbar_kw=dot_hue_cbar_kw,
        matrix_cbar_kw=matrix_cbar_kw,
        no_spines=no_spines,
        no_ticks=no_ticks,
        ax=ax
    )


@doc
def dot(
        dot_size: np.ndarray,
        dot_hue: np.ndarray = None,
        xticklabels: List[str] = None,
        yticklabels: List[str] = None,
        xlabel: str = None,
        ylabel: str = None,
        sizes: Tuple[int, int] = (1, 200),
        dtype: Any = None,
        dot_patch: DotPatch = "circle",
        dot_outline: bool = True,
        outline_color: Any = "#999999",
        alpha: float = 1.0,
        legend_kw: Dict = None,
        no_spines: bool = True,
        no_ticks: bool = True,
        ax: mpl.axes.Axes = None,
) -> DotHeatmap:
    """Dot plot

    Args:
        dot_size: [size]
        dot_hue: [hue]
        xticklabels: [xticklabels]
        yticklabels: [yticklabels]
        xlabel: [xlabel]
        ylabel: [ylabel]
        sizes: [sizes]
        dtype: [dtype]
        dot_patch: The dot will be drawn as "circle" or "pie"
        dot_outline: To draw the dot outline, Default: True
        outline_color: The color of the dot outline, Default: "#999999"
        alpha: Control the opacity
        legend_kw: The options that pass to control legend, should be a dict
        no_spines: [no_spines]
        no_ticks: [no_ticks]
        ax: [ax]

    Returns:
        Return a `DotHeatmap` object

    """

    return DotHeatmap(
        dot_size,
        dot_hue,
        xticklabels=xticklabels,
        yticklabels=yticklabels,
        xlabel=xlabel,
        ylabel=ylabel,
        sizes=sizes,
        dtype=dtype,
        dot_patch=dot_patch,
        outline_color=outline_color,
        alpha=alpha,
        dot_outline=dot_outline,
        dot_size_legend_kw=legend_kw,
        no_spines=no_spines,
        no_ticks=no_ticks,
        ax=ax
    )



from __future__ import annotations

from itertools import cycle
from typing import Union, List, Dict, Callable

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes
from matplotlib.colors import Colormap

from legendkit import CatLegend
from milkviz.utils import doc, set_default
from milkviz.utils import get_cmap_colors, set_spines


def fold_add(arr):
    new_arr = np.zeros(arr.size)
    new_arr[0] = arr[0]
    for i in range(arr.size - 1):
        new_arr[i + 1] = arr[:i + 2].sum()
    return new_arr.astype(arr.dtype)  # preserve the data type


@doc
def stacked_bar(data: pd.DataFrame = None,
                x: str | List | np.ndarray = None,
                y: str | List | np.ndarray = None,
                stacked: Union[str, np.ndarray, None] = None,
                orient: str = "v",
                percentage: bool = False,
                cmap: Union[str, Colormap, None] = None,
                show_values: bool | Callable = False,
                legend_kw: Dict = None,
                ax: mpl.axes.Axes = None,
                **kwargs,
                ) -> Axes:
    """Stacked bar plot

    Args:
        data: [data]
        x: [x]
        y: [y]
        stacked: Which columns to plot as stacked type
        orient: "v" or "h"
        percentage: Normalize value to 1
        cmap: [cmap], default: "echarts", a custom milkviz palette
        show_values: Whether to display values of each block,
            or you can pass in a function to tell when to display
            like `lambda x: x > 100` will only display when value exceed 100
        legend_kw: The options to configure legend
        ax: [ax]
        **kwargs: Pass to `seaborn.barplot <https://seaborn.pydata.org/generated/seaborn.barplot.html#seaborn.barplot>`_

    Returns:
        [return_obj]

    """

    cmap = "echarts" if cmap is None else cmap
    legend_kw = set_default(legend_kw, {})

    show_values_func = lambda a: a
    if isinstance(show_values, Callable):
        show_values_func = show_values
        show_values = True

    if ax is None:
        ax = plt.gca()

    if data is None:
        data = pd.DataFrame({"x": x, "y": y, "stacked": stacked})
        x = "x"
        y = "y"
        stacked = "stacked"
    # This reverse step is to make the label column in the right order
    data = data.iloc[::-1, :]
    value_key, label_key = (y, x) if orient == "v" else (x, y)
    text_key = "display"
    stacked_order = data[stacked].unique()
    label_order = data[label_key].unique()

    all_g = []
    for label in label_order:
        g = data[data[label_key] == label].set_index(stacked)
        s_order = []
        for i in stacked_order:
            if i in g.index:
                s_order.append(i)
        g = g.loc[s_order, :]
        v = g[value_key].to_numpy()
        g[text_key] = v
        if percentage:
            v = v / v.sum()
        g[value_key] = fold_add(v)
        all_g.append(g)
    data_g = pd.concat(all_g).reset_index()

    colors = get_cmap_colors(cmap)
    # This reverse step is to make the stacked column in the right order
    leg_colors, leg_labels = [], []
    for (n, g), c in zip(data_g.iloc[::-1, :].groupby(stacked, sort=False), cycle(colors)):
        bar = sns.barplot(x=x, y=y, data=g, ax=ax, color=c, orient=orient, ci=None, **kwargs)
        if show_values:
            for i in range(len(g)):
                loc = g.iloc[i, :][value_key]
                text = g.iloc[i, :][text_key]
                if show_values_func(float(text)):
                    if orient == "v":
                        bar.text(i, loc, text, ha="center", va="center", bbox=dict(fc="white", alpha=0.7))
                    else:
                        bar.text(loc, i, text, ha="center", va="center", rotation=-90, bbox=dict(fc="white", alpha=0.7))
        leg_colors.append(c)
        leg_labels.append(n)

    legend_kw = set_default(legend_kw, {})
    legend_options = dict(
        handle="square",
        title_align="left",
        bbox_to_anchor=(1.05, 0.5),
        bbox_transform=ax.transAxes,
        loc="center left",
        title=stacked if isinstance(stacked, str) else None,
    )
    legend_options = {**legend_options, **legend_kw}
    CatLegend(leg_colors, leg_labels, ax=ax, **legend_options)
    set_spines(ax, (1, 0, 1, 0))
    return ax



from typing import Union, List, Any, Optional, Dict

import numpy as np
from matplotlib import cm
from matplotlib import colors as mcolors
from matplotlib.colors import Colormap
from matplotlib.colors import to_hex

Colors = Union[str, List[str], Colormap]


def mask_triu(arr, k=1):
    if not isinstance(arr, np.ndarray):
        arr = np.array(arr, dtype=np.float64)
    arr = arr.astype(np.float64)
    shape = arr.shape
    if shape[0] != shape[1]:
        raise ValueError("Must be a square")
    iu = np.triu_indices(shape[0], k=k)
    arr[iu] = np.nan
    return np.fliplr(arr.T)


def norm_arr(arr, size=(1, 100), vmin=None, vmax=None):
    """Normalize the array to a specific range"""
    if not isinstance(arr, np.ndarray):
        arr = np.array(arr)
    arr = arr.flatten()
    add_count = 0
    if vmin is not None:
        arr = np.array([*arr, vmin])
        add_count += 1
    if vmax is not None:
        arr = np.array([*arr, vmax])
        add_count += 1
    amin = np.nanmin(arr)
    amax = np.nanmax(arr)
    size_range = size[1] - size[0]
    arr = (arr - amin) * (size_range / (amax - amin))
    arr += size[0]
    slice_ = -add_count if add_count > 0 else None
    return arr[:slice_]


def adaptive_figsize(size, min_side=5):
    """scale the min number to min_side, while maintain the ratio"""
    size = np.array(size, dtype=np.float64)
    amin = np.amin(size)
    min_ix = np.argmin(size)
    scale = min_side / amin
    if amin < min_side:
        size *= scale
        size[min_ix] = min_side
    return tuple(size)


def get_ax_size(ax):
    x = ax.get_xlim()
    y = ax.get_ylim()
    x_side = x[1] - x[0]
    y_side = y[1] - y[0]
    return x_side, y_side


def get_render_size(ax):
    f = ax.get_figure()
    return f.get_figwidth(), f.get_figheight()


def set_spines(ax, status=(0, 0, 0, 0)):
    for spine, s in zip(ax.spines.values(), status):
        spine.set_visible(s)


def set_ticks(ax, xticks="none", yticks="none"):
    ax.xaxis.set_ticks_position(xticks)
    ax.yaxis.set_ticks_position(yticks)


def create_cmap(c_array: List[str], name: str = "custom_cmap") -> Colormap:
    return mcolors.ListedColormap(c_array, name=name)


def get_cmap_colors(color: Colors):
    """This utility function allow different color input
    
    1) colormap name
    2) colormap instance
    3) color name, list of color name
    
    """
    cmap = None
    if isinstance(color, str):
        cmap = cm.get_cmap(color)
    elif isinstance(color, Colormap):
        cmap = color
    else:
        return [to_hex(c, keep_alpha=True) for c in color]

    return [to_hex(cmap(i), keep_alpha=True) for i in range(cmap.N)]


def color_mapper_cat(types: Union[List[Any], np.ndarray],
                     c_array: Optional[List[str]] = None,
                     cmap: Colors = None,
                     ) -> Dict:
    """
    
    Args:
        types: All input types
        c_array: The color array that maps to the types
        cmap: An array of colors or colormap


    Returns:
        {type: color}
        
    """
    N = len(types)
    types = np.array(types)
    _, idx = np.unique(types, return_index=True)
    uni_types = types[np.sort(idx)]
    if c_array is not None:
        if len(c_array) < N:
            raise ValueError(f"The length of input color array {len(c_array)} does not match types {N}")
        else:
            cmapper = dict(zip(types, c_array))
            return {k: cmapper[k] for k in uni_types}
    if cmap is not None:
        all_colors = []
        colors_pool = get_cmap_colors(cmap)
        while len(all_colors) < N:
            all_colors += colors_pool

        return dict(zip(uni_types, all_colors))


def color_mapper_val(values: Union[List[Any], np.ndarray],
                     c_array: Optional[List[str]] = None,
                     cmap: Colors = None, ) -> List[str]:
    if not isinstance(values, np.ndarray):
        values = np.ndarray(values)
    vmin = np.nanmin(values)
    vmax = np.nanmax(values)
    norm = mcolors.Normalize(vmin=vmin, vmax=vmax)
    if c_array is not None:
        cmap = create_cmap(c_array)
    mapper = cm.ScalarMappable(norm=norm, cmap=cmap)
    return [to_hex(mapper.to_rgba(v), keep_alpha=True) for v in values]



from __future__ import annotations

from typing import Optional, List, Tuple, Any, Dict

import matplotlib as mpl
import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from legendkit import SizeLegend, Colorbar
from milkviz.utils import norm_arr, doc, set_default, set_spines


@doc
def bubble(data: Optional[pd.DataFrame] = None,
           x: str | List | np.ndarray = None,
           y: str | List | np.ndarray = None,
           hue: str | List | np.ndarray = None,
           size: str | List | np.ndarray = None,
           cmap: Any = None,
           sizes: Tuple[int, int] = (10, 250),
           dtype: Any = None,
           legend_kw: Dict = None,
           cbar_kw: Dict = None,
           ax: mpl.axes.Axes = None,
           ) -> mpl.axes.Axes:
    """Bubble plot

    Args:
        data: [data]
        x: [x]
        y: [y]
        hue: [hue]
        size: [size]
        cmap: [cmap], default to "RdBu"
        sizes: [sizes]
        dtype: [dtype]
        legend_kw: The options to configure legend
        cbar_kw: The options to configure colorbar
        ax: [ax]

    Returns:
        [return_obj]

    """
    cmap = set_default(cmap, "RdBu")
    ax = set_default(ax, plt.gca())
    legend_kw = set_default(legend_kw, {})
    cbar_kw = set_default(cbar_kw, {})

    if data is not None:
        x = data[x].to_numpy()
        y = data[y].to_numpy()
        hue = data[hue].to_numpy() if hue is not None else None
        size = data[size].to_numpy() if size is not None else None

    if size is None:
        raise ValueError("At least `size` must be provided")

    circ_size = norm_arr(size, sizes)
    bubbles = ax.scatter(x, y, s=circ_size, c=hue, cmap=cmap)

    legend_options = dict(
        bbox_to_anchor=(1.05, 1),
        bbox_transform=ax.transAxes,
        loc="upper left",
        title_align="left",
        dtype=dtype
    )
    legend_options = {**legend_options, **legend_kw}
    SizeLegend(circ_size, array=size, ax=ax, **legend_options)

    if hue is not None:
        cbar_options = dict(
            bbox_to_anchor=(1.05, 0),
            bbox_transform=ax.transAxes,
            loc="lower left",
            title_align="left",
        )
        cbar_options = {**cbar_options, **cbar_kw}
        Colorbar(bubbles, ax=ax, **cbar_options)
    set_spines(ax, (1, 0, 1, 0))

    return ax



import re
from inspect import getdoc

PARAMETERS_DOCSTRINGS = dict(
    ax="Pre-existing axes for the plot. Or a new one will be created.",
    data="Data use to plot, should be a dataframe",
    x="Either a key in data or a numpy array",
    y="Either a key in data or a numpy array",
    hue="The color array that map to marker colors",
    size="The size array that map to marker sizes",
    cmap="The colormap to be used, either a cmap name or a `matplotlib.cm.ColorMap`",
    xticklabels="Text to put on the x-axis ticks",
    yticklabels="Text to put on the y-axis ticks",
    xlabel="Title for x-axis",
    ylabel="Title for y-axis",
    types="The types array that tells the types",
    values="The values array that tells the values",
    vminmax="Remap your values to another range",
    sizes="The size range of circle markers, (min, max)",
    no_spines="If True, will turn off the frame of the plot",
    no_ticks="If True, will turn off both major and minor ticks",
    legend="If True, show legend",
    legend_title="The title of legend",
    legend_pos="A 2-tuple (x, y) to set the position of legend, the origin is at the upper left",
    legend_ncol="Number of columns the legend will have",
    dtype="Only coerce the value display on the legend, does not affect plotting",

    return_obj="A `matplotlib.axes.Axes` instance",
    cbar_title="The title of colorbar",
    cbar_pos="A 2-tuple (x, y) or 4-tuple (x, y, frac of fig width, frac of fig height) "
             "to place colorbar, the origin is the same as plot",
    cbar_size="A tuple (width, height) in inches",
    cbar_ticklabels="Set the ticklabels of colorbar"
)


def doc(obj):
    docstring = getdoc(obj)
    for param_name, content in PARAMETERS_DOCSTRINGS.items():
        pattern = re.compile(f"""(\[{param_name}\])""")
        docstring = re.sub(pattern, content, docstring)
    obj.__doc__ = docstring
    return obj



from __future__ import annotations

from typing import List, Optional, Dict, Any

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import cm

from legendkit import CatLegend, Colorbar, vstack, hstack
from milkviz.utils import color_mapper_cat, doc, set_default


@doc
def anno_clustermap(
        data: pd.DataFrame,
        # define row colors
        row_colors: str | List[str] = None,
        row_colors_cmap: str | List[str] = None,
        row_colors_order: Dict = None,
        row_colors_label: Dict = None,
        row_label: str = None,
        row_legend_title: str = None,
        row_legend_split: bool = True,
        # define col colors
        col_colors: str | List[str] = None,
        col_colors_cmap: str | List[str] = None,
        col_colors_order: Dict = None,
        col_colors_label: Dict = None,
        col_label: str = None,
        col_legend_title: str = None,
        col_legend_split: bool = True,
        heat_cmap: Any = None,
        legend_padding: float = 5,
        legend_kw: Dict = None,
        cbar_kw: Dict = None,
        categorical_cbar: Optional[List[str]] = None,
        cbar_title: str = None,
        **kwargs,
) -> sns.matrix.ClusterGrid:
    """Color or label annotated clustermap

    Args:
        data: [data], multi-levels annotations should store in MultiIndex
        {row|col}_label: The index/columns level used for x-axis/y-axis label
        {row|col}_colors: The index/columns level used to label in color stripe
        row_colors_cmap: The colormap for row_colors, default: "tab20"
        col_colors_cmap: The colormap for col_colors, default: "echarts", a custom colormap in milkviz
        {row|col}_colors_order: A dict-like mapper to specific the order in each level,
        {row|col}_colors_label: A dict-like mapper to overwrite the name of each level,
        {row|col}_legend_split: Whether to split each level of colors stripe
        {row|col}_legend_title: The title of row legend, when row_legend_split = False
        legend_padding: The space between legend
        legend_kw: Options to customize legend, will be applied to all,
        cbar_kw: Options to customize colorbar, if categorical cbar, use legend_kw
        heat_cmap: The colormap for heatmap, default: "RdBu_r"
        categorical_cbar: Turn the colorbar in to categorical legend in text
        cbar_title: Set the title for colorbar
        **kwargs: Pass to `seaborn.clustermap <https://seaborn.pydata.org/generated/seaborn.clustermap.html#seaborn.clustermap>`_

    Returns:
        A `seaborn.matrix.ClusterGrid` instance

    """
    row_colors_cmap = "tab20" if row_colors_cmap is None else row_colors_cmap
    col_colors_cmap = "echarts" if col_colors_cmap is None else col_colors_cmap
    heat_cmap = "RdBu_r" if heat_cmap is None else heat_cmap
    legend_kw = set_default(legend_kw, {})
    cbar_kw = set_default(cbar_kw, {})

    # split the dataframe
    data = data.copy()
    raw_data = data.to_numpy()

    if isinstance(row_colors, str):
        row_colors = [row_colors]

    if isinstance(col_colors, str):
        col_colors = [col_colors]

    if (row_colors is not None) & (row_colors_order is not None):
        # data.reorder_levels(list(row_colors_order.keys()), axis=0)
        for name, labels in row_colors_order.items():
            data = data.reindex(axis="index", level=name, labels=labels)

    if (col_colors is not None) & (col_colors_order is not None):
        # data.reorder_levels(list(col_colors_order.keys()), axis=1)
        for name, labels in col_colors_order.items():
            data = data.reindex(axis="column", level=name, labels=labels)

    if (row_colors is not None) & (row_colors_label is not None):
        data = data.rename_axis(index=row_colors_label)
        row_colors = [row_colors_label[label] for label in row_colors]

    if (col_colors is not None) & (col_colors_label is not None):
        data = data.rename_axis(columns=col_colors_label)
        col_colors = [col_colors_label[label] for label in col_colors]

    row_info = data.index.to_frame(index=False)
    col_info = data.columns.to_frame(index=False)
    row_label = row_info[row_label] if row_label is not None else None
    col_label = col_info[col_label] if col_label is not None else None
    plot_data = pd.DataFrame(raw_data, index=row_label, columns=col_label)

    clustermap_kwargs = dict(cmap=heat_cmap, cbar_pos=None, **kwargs)
    row_colors_mapper = None
    col_colors_mapper = None
    row_legend_contents = None
    col_legend_contents = None

    if row_colors is not None:
        info = row_info[row_colors]
        row_legend_contents = {rc: np.unique(info[rc].to_numpy().flatten()) for rc in row_colors}
        row_colors_mapper = color_mapper_cat(info.to_numpy().flatten(), cmap=row_colors_cmap)
        clustermap_kwargs["row_colors"] = info.replace(row_colors_mapper)

    if col_colors is not None:
        info = col_info[col_colors]
        col_legend_contents = {cc: np.unique(info[cc].to_numpy().flatten()) for cc in col_colors}
        col_colors_mapper = color_mapper_cat(info.to_numpy().flatten(), cmap=col_colors_cmap)
        clustermap_kwargs["col_colors"] = info.replace(col_colors_mapper)

    g = sns.clustermap(plot_data, **clustermap_kwargs)

    # plot row colors legend
    legend_transform = dict(
        bbox_to_anchor=(1.05, 1),
        bbox_transform=g.ax_heatmap.transAxes,
        loc="upper left",
    )

    cbar_transform = dict(
        bbox_to_anchor=(1.05, 0),
        ax=g.ax_heatmap,
        bbox_transform=g.ax_heatmap.transAxes,
        loc="lower left",
    )

    legend_options = dict(
        handle="square",
        title_align="left",
    )
    legend_options = {**legend_options, **legend_kw}

    compose_legends = []
    # plot col colors legend
    if col_colors is not None:
        if col_legend_split:
            legend_list = []
            for k, labels in col_legend_contents.items():
                leg = CatLegend(
                    colors=[col_colors_mapper[i] for i in labels],
                    labels=labels,
                    title=k,
                    **legend_options
                )
                legend_list.append(leg)
            g.col_color_legend = hstack(legend_list, spacing=legend_padding)
        else:
            title = col_colors[0] if col_legend_title is None else col_legend_title
            labels, colors = zip(*col_colors_mapper.items())
            g.col_color_legend = CatLegend(colors, labels, ax=g.ax_heatmap, title=title, **legend_options)
        compose_legends.append(g.col_color_legend)

    if row_colors is not None:
        if row_legend_split:
            legend_list = []
            for k, labels in row_legend_contents.items():
                leg = CatLegend(
                    colors=[row_colors_mapper[i] for i in labels],
                    labels=labels,
                    title=k,
                    **legend_options
                )
                legend_list.append(leg)
            g.row_color_legend = hstack(legend_list, spacing=legend_padding)
        else:
            title = row_colors[0] if row_legend_title is None else row_legend_title
            labels, colors = zip(*row_colors_mapper.items())
            g.row_color_legend = CatLegend(colors, labels, title=title, title_align="left")
        compose_legends.append(g.row_color_legend)

    if categorical_cbar is not None:
        cmap = cm.get_cmap(heat_cmap)
        cmap_size = np.linspace(0, cmap.N, len(categorical_cbar), dtype=int)
        labels, colors = categorical_cbar, [cmap(i) for i in cmap_size]
        g.cbar = CatLegend(colors, labels, title=cbar_title, **cbar_transform, **legend_options)
        compose_legends.append(g.cbar)
    else:
        cdata = g.data2d.to_numpy()
        cmin = np.nanmin(cdata)
        cmax = np.nanmax(cdata)
        cbar_options = dict(
            orientation="horizontal",
            title_align="left",
        )
        cbar_options = {**cbar_options, **cbar_kw}
        g.cbar = Colorbar(vmin=cmin, vmax=cmax, cmap=heat_cmap, title=cbar_title, **cbar_transform, **cbar_options)

    g.compose_legends = vstack(compose_legends, padding=0, spacing=legend_padding, align="left", **legend_transform)
    g.ax_heatmap.add_artist(g.compose_legends)

    # close labels and ticks on y-axis
    if row_label is None:
        g.ax_heatmap.set_yticklabels("")
        g.ax_heatmap.set_ylabel("")
        g.ax_heatmap.set_yticks([])

    # close labels and ticks on x-axis
    if col_label is None:
        g.ax_heatmap.set_xticklabels("")
        g.ax_heatmap.set_xlabel("")
        g.ax_heatmap.set_xticks([])

    return g



from collections import Counter
from typing import Optional, Union, Set, List, Any

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes

from milkviz.utils import doc


def find_intersec(*arr_list):
    count = 0
    counter = [Counter(arr) for arr in arr_list]
    max_ix = np.argmax([len(c) for c in counter])
    max_size_counter = counter[max_ix]
    counter.pop(max_ix)
    for k1 in max_size_counter.keys():
        count += min(max_size_counter[k1], *[c.get(k1, 0) for c in counter])
    return count


@doc
def venn(
        data: Union[List[Set], List[List[Any]], List[int]],
        names: Optional[List[str]] = None,
        colors: Optional[List[str]] = None,
        alpha: Union[float, List[float], None] = None,
        subset_labels: Optional[List[str]] = None,
        normalize_to: float = 1.0,
        weighted: bool = True,
        ax: Optional[Axes] = None,
) -> Axes:
    """Venn diagrame for 2 & 3 Sets

    This is a wrapper around `matplotlib-venn <https://github.com/konstantint/matplotlib-venn>`_,
    allows you to input raw data and compute intersection for you.

    Args:
        data: 1) A list of sets 2) A list of list, will compute intersection, duplication
            will be considered. 3) A list of number denotes the regions in venn diagram
            in the following order: (10, 01, 11) or (100, 010, 110, 001, 101, 011, 111)
        names: The name for each sets
        colors: The color for each sets
        alpha: The alpha value for each sets
        subset_labels: The label for each small regions, usually denotes the size, following the order
            as described before
        normalize_to: specifies the total (on-axes) area of the circles to be drawn.
            Sometimes tuning it (together with the overall fiture size) may be useful
            to fit the text labels better.
        weighted: If True, the circle area is weighted by the size of sets
        ax: [ax]

    Returns:
        [return_obj]

    """
    try:
        from matplotlib_venn import venn2, venn3, venn2_unweighted, venn3_unweighted
    except ImportError:
        raise ImportError("Extra dependencies needed, Try `pip install matplotlib-venn`")
    if ax is None:
        ax = plt.gca()
    s = len(data)
    venn_func = None
    patch_order = None
    colors = ("#F1C266", "#65BBC4", "#F3A292") if colors is None else colors
    venn_num = 0
    # compute intersection for user
    if not isinstance(data[0], (Set, int)):
        if s == 2:
            s1, s2 = data
            s1s2 = find_intersec(s1, s2)
            s1, s2 = len(s1), len(s2)
            data = (s1 - s1s2, s2 - s1s2, s1s2)
            venn_num = 2
        else:
            s1, s2, s3 = data
            s1s2, s1s3, s2s3, s1s2s3 = find_intersec(s1, s2), find_intersec(s1, s3), \
                                       find_intersec(s2, s3), find_intersec(s1, s2, s3)
            s1, s2, s3 = len(s1), len(s2), len(s3)
            data = (s1 - s1s2 - s1s3 + s1s2s3, s2 - s1s2 - s2s3 + s1s2s3, s1s2 - s1s2s3,
                    s3 - s1s3 - s2s3 + s1s2s3, s1s3 - s1s2s3, s2s3 - s1s2s3, s1s2s3)
            venn_num = 3
    # user input is regions area
    elif isinstance(data[0], int):
        if s == 3:
            venn_num = 2
        elif s == 7:
            venn_num = 3
        else:
            raise ValueError(f"You have {s} input. Does not match ordering of"
                             f"(10, 01, 11) or (100, 010, 110, 001, 101, 011, 111)")
    # user input is sets
    else:
        if s > 3:
            raise ValueError(f"You have {s} sets. Venn diagram is usually for "
                             f"showing intersection between no more than 3 sets,"
                             f"please use upset plot instead for sets > 3.")
        venn_num = s

    if venn_num == 2:
        venn_func = venn2 if weighted else venn2_unweighted
        patch_order = ('10', '01', '11')
    elif venn_num == 3:
        venn_func = venn3 if weighted else venn3_unweighted
        patch_order = ('100', '010', '110', '001', '101', '011', '111')
    else:
        raise ValueError(f"Intersection only happens between more than 2 sets.")

    v = venn_func(data, names, colors, normalize_to=normalize_to, ax=ax)
    alpah_array = [0.4 for _ in range(len(patch_order))]
    if alpha is not None:
        if isinstance(alpha, (int, float)):
            alpah_array = [alpha for _ in range(len(patch_order))]
        else:
            if len(alpha) >= s:
                alpah_array = alpha

    for patch_id, a in zip(patch_order, alpah_array):
        p = v.get_patch_by_id(patch_id)
        if p is not None:
            p.set_alpha(a)

    if subset_labels is not None:
        for patch_id, label in zip(patch_order, subset_labels):
            l = v.get_label_by_id(patch_id)
            if l is not None:
                l.set_text(label)

    return ax



from .doc import doc
from .fig import adaptive_figsize, norm_arr, \
    set_ticks, set_spines, color_mapper_cat, \
    color_mapper_val, mask_triu, get_cmap_colors, \
    create_cmap, get_render_size
from .geo import rotate_points, normalize
from .params import set_default



from matplotlib import cm

from ._bubble import bubble
from ._cell_map import point_map, polygon_map, point_map3d
from ._clustermap import anno_clustermap
# from ._dot import dot
from ._dot_matrix import dot, dot_heatmap
from ._graph import graph
from ._stacked_bar import stacked_bar
from ._upset import upset
from ._venn import venn
from .colormap import echarts, tailwind, retro_metro, river_nights, dutch_field, spring_pastels
from .utils import mask_triu

cm.register_cmap(cmap=echarts)
cm.register_cmap(cmap=tailwind)
cm.register_cmap(cmap=retro_metro)
cm.register_cmap(cmap=dutch_field)
cm.register_cmap(cmap=river_nights)
cm.register_cmap(cmap=spring_pastels)
